/*  ht2992.h -- Header file containing register definiitions,
    etc. for Hytec 2992 Serial Highway Link Driver Module.

    Author: Marty Wise
    Origination: July 29, 1993

    $Log: ht2992.h,v $
 * Revision 1.1  1994/10/20  20:18:09  tang
 * Commit adding of Camac Driver HighTech h file.
 *
 * Revision 1.11  94/05/11  15:51:21  15:51:21  wise (Marty Wise)
 * Removed check for bogus crate addresses used for debug
 * 
 * Revision 1.10  94/05/11  14:05:13  14:05:13  wise (Marty Wise)
 * Added low-level diagnostics
 * 
 * Revision 1.9  94/05/09  10:21:59  10:21:59  wise (Marty Wise)
 * Updated CSSA to be similar to CFSA. Previously was lacking diagnostics, etc.
 * 
 * Revision 1.8  94/05/09  09:42:00  09:42:00  wise (Marty Wise)
 * Added code to test for SHR and XMR prior to calling command complete.
 * This was for workaround for HyTec PLD bug.
 * 
 * Revision 1.7  94/04/11  14:14:33  14:14:33  wise (Marty Wise)
 * no changes
 * 
 * Revision 1.6  94/03/29  07:36:30  07:36:30  wise (Marty Wise)
 * Reduced TIMERVAL by factor of 100
 * 
 * Revision 1.5  94/03/29  07:31:04  07:31:04  wise (Marty Wise)
 * Corrected error in ht2992_rpt, was 0x0400, should have been 0x4000. This error resulted in 
 * block read functions returning only the first data point multiple times since, new reads
 * (repeat) were not initiated.
 * 
 * Revision 1.4  94/03/27  07:26:02  07:26:02  wise (Marty Wise)
 * Added Q check on unbypass
 * 
 * Revision 1.3  94/02/11  10:54:18  10:54:18  wise (Marty Wise)
 * Reduced timerval by factor of 100, should produce timeout on the order of
 * 10 msec.
 * 
 * Revision 1.2  94/01/14  16:02:48  16:02:48  wise (Marty Wise)
 * Added automatic insertion into loop
 * 
 * Revision 1.1  93/11/05  06:39:55  06:39:55  wise (Marty Wise)
 * Initial revision
 * 
 * Revision 0.7  93/10/21  16:53:13  16:53:13  wise (Marty Wise)
 * Added preliminary version of csubc, csubr, and csgfa
 * 
 * Revision 0.6  93/10/21  07:25:50  07:25:50  wise (Marty Wise)
 * Added CFUBBR_pre, CFUBR_main, CFUBR_post and CFUBR
 * 
 * Revision 0.5  93/10/19  09:03:58  09:03:58  wise (Marty Wise)
 * added cssa and csmad.
 * 
 * Revision 0.4  93/10/12  07:33:20  07:33:20  wise (Marty Wise)
 * cfubc working, cfmad working, almost ready for alpha release.
 * 
 * Revision 0.3  93/10/10  10:12:06  10:12:06  wise (Marty Wise)
 * Cleaned up macro implementation. Split CFSA into macros (CFSA_pre, CFSA_main
 * and CFSA_post) as well as cover macro (CFSA). This allows multiple action
 * functions to remove overhead from tight loops.
 * 
 * Revision 0.2  93/10/08  08:09:14  08:09:14  wise (Marty Wise)
 * Checked in with rcs identifiers
 * 
*/
static	char	ht2992_rcsid[] = "$Header: /net/phoebus/epicsmgr/cvsroot/epics/base/include/ht2992.h,v 1.1 1994/10/20 20:18:09 tang Exp $" ;


#define	ht2992_rdid	*(unsigned short *) 0xffffffc0
#define	ht2992_rmod	*(unsigned short *) 0xffffffc2
#define	ht2992_rcsr	*(unsigned short *) 0xffffffc4
#define	ht2992_rlmk	*(unsigned short *) 0xffffffca
#define	ht2992_rvec	*(unsigned short *) 0xffffffcc
#define ht2992_rlst	*(unsigned short *) 0xffffffce
#define	ht2992_rsts	*(unsigned short *) 0xffffffd6
#define ht2992_rdst	*(unsigned short *) 0xffffffd8
#define ht2992_rddb	*(unsigned short *) 0xffffffda
#define	ht2992_rddt	*(unsigned short *) 0xffffffdc
#define ht2992_rdip	*(unsigned short *) 0xffffffde
#define ht2992_rdpt	*(unsigned short *) 0xffffffe0
#define ht2992_rups	*(unsigned short *) 0xffffffe2
#define	ht2992_rtcs	*(unsigned short *) 0xfffffff0
#define	ht2992_rtxn	*(unsigned short *) 0xfffffff2
#define	ht2992_rtxa	*(unsigned short *) 0xfffffff4
#define	ht2992_rtxf	*(unsigned short *) 0xfffffff6
#define	ht2992_rdt1	*(unsigned short *) 0xfffffff8
#define	ht2992_rdt2	*(unsigned short *) 0xfffffffa
#define	ht2992_rdt3	*(unsigned short *) 0xfffffffc
#define	ht2992_rdt4	*(unsigned short *) 0xfffffffe


#define	ht2992_wvec	*(unsigned short *) 0xffffffc0
#define	ht2992_wien	*(unsigned short *) 0xffffffc4
#define	ht2992_wlmk	*(unsigned short *) 0xffffffca
#define	ht2992_wclm	*(unsigned short *) 0xffffffcc
#define	ht2992_wscl	*(unsigned short *) 0xffffffce
#define	ht2992_wcll	*(unsigned short *) 0xffffffd0
#define	ht2992_wscm	*(unsigned short *) 0xffffffd2
#define	ht2992_wssm	*(unsigned short *) 0xffffffd4
#define	ht2992_wnaf	*(unsigned short *) 0xffffffd6
#define	ht2992_wtca	*(unsigned short *) 0xffffffd8
#define	ht2992_wtdb	*(unsigned short *) 0xffffffda
#define	ht2992_wtdt	*(unsigned short *) 0xffffffdc
#define ht2992_wdsf	*(unsigned short *) 0xffffffde
#define	ht2992_wdsp	*(unsigned short *) 0xffffffe0
#define	ht2992_wups	*(unsigned short *) 0xffffffe2
#define	ht2992_wfif	*(unsigned short *) 0xffffffe4
#define	ht2992_wlbf	*(unsigned short *) 0xffffffe6


/************************************************/
/* CSR Register bit defines			*/
#define ht2992_rcsr_busy	((ht2992_rcsr & 0x0001) >> 0)
#define ht2992_rcsr_rdy		((ht2992_rcsr & 0x0010) >> 4)
#define ht2992_rcsr_xrsp	((ht2992_rcsr & 0x0020) >> 5)
#define ht2992_rcsr_qrsp	((ht2992_rcsr & 0x0040) >> 6)
#define ht2992_rcsr_oies	((ht2992_rcsr & 0x0080) >> 7)
#define ht2992_rcsr_xmr		((ht2992_rcsr & 0x0100) >> 8)
#define ht2992_rcsr_shr		(ht2992_rcsr & 0x0200)
#define ht2992_rcsr_lam		((ht2992_rcsr & 0x0400) >> 10)
#define ht2992_rcsr_sync	((ht2992_rcsr & 0x0800) >> 11)
#define ht2992_rcsr_dmd		((ht2992_rcsr & 0x1000) >> 12)
#define ht2992_rcsr_done	(ht2992_rcsr & 0x0300)


/************************************************/
/* LAM Status Register bit definitions		*/
#define ht2992_rlst_sq		((ht2992_rlst & 0x0001) >> 0)
#define ht2992_rlst_sx		((ht2992_rlst & 0x0002) >> 1)
#define ht2992_rlst_err		((ht2992_rlst & 0x0004) >> 2)
#define ht2992_rlst_derr	((ht2992_rlst & 0x0008) >> 3)
#define ht2992_rlst_adnr	((ht2992_rlst & 0x0010) >> 4)
#define ht2992_rlst_lpe		((ht2992_rlst & 0x0020) >> 5)
#define ht2992_rlst_tmo		((ht2992_rlst & 0x0040) >> 6)
#define ht2992_rlst_tpe		((ht2992_rlst & 0x0080) >> 7)
#define ht2992_rlst_rdy		((ht2992_rlst & 0x0100) >> 8)
#define ht2992_rlst_xmr		((ht2992_rlst & 0x0200) >> 9)
#define ht2992_rlst_dmd		((ht2992_rlst & 0x0400) >> 10)
#define ht2992_rlst_serr	((ht2992_rlst & 0x0800) >> 11)
#define ht2992_rlst_esum	((ht2992_rlst & 0x08fc) >> 0)
#define ht2992_rlst_ok		(ht2992_rlst ^ 0x0cfc)

/************************************************/
/* LAM Mask Definitions				*/
#define	MASK_sq			0x0001
#define MASK_sx			0x0002
#define	MASK_err		0x0004
#define MASK_derr		0x0008
#define	MASK_adnr		0x0010
#define	MASK_lpe		0x0020
#define	MASK_tmo		0x0040
#define	MASK_tpe		0x0080
#define	MASK_rdy		0x0100
#define	MASK_xmr		0x0200
#define	MASK_dmd		0x0400
#define	MASK_serr		0x0800



/************************************************/
/* Full Status register bit definitions		*/
#define ht2992_rsts_err		((ht2992_rsts & 0x0001) >> 0)
#define ht2992_rsts_sx		((ht2992_rsts & 0x0002) >> 1)
#define ht2992_rsts_sq		((ht2992_rsts & 0x0004) >> 2)
#define ht2992_rsts_derr	((ht2992_rsts & 0x0008) >> 3)
#define ht2992_rsts_m1m2	((ht2992_rsts & 0x0030) >> 4)
#define ht2992_rsts_ca		((ht2992_rsts & 0x0fc0) >> 6)


#define naf_c_mask	0xffff0000
#define naf_c_shift	16
#define	naf_f_mask	0x001f
#define	naf_f_shft	0
#define	naf_a_mask	0x01e0
#define	naf_a_shft	5
#define	naf_n_mask	0x3e00
#define	naf_n_shft	9

#define ht2992_rpt_flag	0x4000

#define TIMERVAL	10000

#define SPINWAIT							\
	{								\
		register unsigned short *WAIT_rcsr;			\
		int	WAIT_COUNT;					\
									\
		WAIT_rcsr = (unsigned short *)&ht2992_rcsr;		\
		if (ht2992_rcsr_done != 0x3)				\
			for (WAIT_COUNT = 0; 				\
			     ((WAIT_COUNT < TIMERVAL) && 		\
			      ((*WAIT_rcsr & 0x0300) != 0x0300));	\
			     WAIT_COUNT++)				\
				;					\
	}

#define SPINLOCK							\
	{								\
		register unsigned short *LOCK_rcsr_ok;			\
		int	LOCK_COUNT;					\
									\
		while (ht2992_rcsr_busy)				\
			;						\
									\
		LOCK_rcsr_ok = (unsigned short *)&ht2992_rcsr;		\
		if (!ht2992_rcsr_shr)					\
			for (LOCK_COUNT = 0; 				\
			     ((LOCK_COUNT < TIMERVAL) && 		\
			      (!(*LOCK_rcsr_ok & 0x0200))); 		\
			     LOCK_COUNT++)				\
				;					\
	}


#define CDREG(CDREG_ext, CDREG_branch, CDREG_crate, CDREG_slot, CDREG_addr)\
{									\
    unsigned int	CDREG_temp	= 0;				\
    unsigned int	CDREG_ctrlr	= 0;				\
    int		ctrlr_data	= 0x100;				\
    int		ctrlr_q		= 0;					\
									\
    /************************************************************/	\
    /* If branch != 0 ==> Error, return with ext = 0		*/	\
    /************************************************************/	\
    if (CDREG_branch != 0)	/* only branch 0 supported 	*/	\
    {									\
	printf("Attempt to declare channel on unsupported branch: %d\n",\
		CDREG_branch);						\
	*CDREG_ext = 0;							\
	return;								\
    }									\
									\
    /************************************************************/	\
    /* If branch has not been initialized, return with ext = 0	*/	\
    /************************************************************/	\
    if (!link_status.c_link)						\
    {									\
	printf("Attempt to declare channel on uninitialize branch\n");	\
	*CDREG_ext = 0;							\
	return;								\
    }									\
									\
    /************************************************************/	\
    /* If parameters out of range, return with *ext = 0		*/	\
    /************************************************************/	\
    if ((CDREG_crate < 0) || (CDREG_crate >= 64) || 			\
	(CDREG_slot < 1) || (CDREG_slot > 30) || 			\
	(CDREG_addr < 0) || (CDREG_addr >= 16))				\
    {									\
	printf("Invalid parameters to CDREG: \n");			\
	printf("\tcrate      = %d\n", CDREG_crate);			\
	printf("\tslot       = %d\n", CDREG_slot);			\
	printf("\tsubaddress = %d\n", CDREG_addr);			\
	*CDREG_ext = 0;							\
	return;								\
    }									\
									\
    /************************************************************/	\
    /* If this crate controller has not been initialized, do so	*/	\
    /************************************************************/	\
    if (!link_status.ctrlr[CDREG_crate])				\
    {									\
	/* pack crate, slot and addr of ctrlr into temp var */		\
	CDREG_ctrlr |= ((CDREG_crate << naf_c_shift) & naf_c_mask);	\
	CDREG_ctrlr |= ((30 << naf_n_shft) & naf_n_mask);		\
	CDREG_ctrlr |= ((0 << naf_a_shft) & naf_a_mask);		\
									\
	/* Un-Bypass crate controller 			*/		\
	cfsa(17, CDREG_ctrlr, &ctrlr_data, &ctrlr_q);			\
	if (!ctrlr_q)							\
		printf("no Q returned from ctrlr unbypass\n");		\
									\
	/* Clear dataway I				*/		\
	ccci(CDREG_ctrlr, 0);						\
									\
	/* Mark controller as initialized to avoid this later	*/	\
	link_status.ctrlr[CDREG_crate] = 1;				\
    }									\
									\
    /************************************************************/	\
    /* pack crate, slot and addr into temp			*/	\
    /************************************************************/	\
    CDREG_temp |= ((CDREG_crate << naf_c_shift) & naf_c_mask);		\
    CDREG_temp |= ((CDREG_slot << naf_n_shft) & naf_n_mask);		\
    CDREG_temp |= ((CDREG_addr << naf_a_shft) & naf_a_mask);		\
									\
    *CDREG_ext = CDREG_temp;						\
								    	\
}									\



#define		CFSA_pre(CFSA_f, CFSA_camac_lock)			\
{									\
									\
	if ((CFSA_f < 0) || (CFSA_f >= 32))				\
	{								\
		return;							\
	}								\
	else								\
	{								\
		/* Lock out interrupts in critical section 	*/	\
		CFSA_camac_lock = intLock();				\
	}								\
									\
}									\


#define		CFSA_main(CFSA_f, CFSA_ext, CFSA_dat, CFSA_q)		\
{									\
    unsigned short	CFSA_temp, CFSA_temp1, CFSA_temp2;		\
    int	CFSA_try, CFSA_read_try, CFSA_write_try, CFSA_cmd_try;		\
    int	MAX_TRIES	= 3;						\
    int	b_temp, c_temp, n_temp, a_temp;					\
									\
    /* Setup crate address register				*/	\
    ht2992_wtca = *((short *)(&CFSA_ext));				\
							    		\
    /* Setup NAF (commit for RD/CMD)				*/	\
    ht2992_wnaf = (*((short *)(&CFSA_ext) + 1) | CFSA_f);		\
							    		\
    /* If this is a write (xxx10xxx pattern)			*/	\
    if ((CFSA_f & 0x0018) == 0x0010)					\
    {									\
	ht2992_wtdt = *(short *)(CFSA_dat);				\
	ht2992_wtdb = *((short *)(CFSA_dat) + 1);			\
    }									\
							    		\
    /* Spin on status bit until reply or timeout received	*/	\
    SPINWAIT								\
									\
									\
    /* If timeout or other error set, clear error flag		*/	\
    /*     and do "re-read" of L2 controller			*/	\
    CFSA_try = CFSA_read_try = CFSA_write_try = CFSA_cmd_try = 0;	\
    while ((ht2992_rcsr & 0x0400) && (CFSA_try < MAX_TRIES))		\
    {									\
	if (debug_hook)							\
	{								\
	    if (ht2992_rlst_sq == 1)					\
	    {								\
		debug_dat.lam_count[0]++;				\
		if (debug_hook > 1)					\
		{							\
		    printf("No Q returned from last CFSA\n\t");		\
		    print_reg(CFSA_ext);				\
		}							\
	    }								\
	    if (ht2992_rlst_sx)						\
	    {								\
		debug_dat.lam_count[1]++;				\
		if (debug_hook > 1)					\
		{							\
		    printf("No X Returned from last CFSA\n");		\
		    print_reg(CFSA_ext);				\
		}							\
	    }								\
	    if (ht2992_rlst_err)					\
	    {								\
		debug_dat.lam_count[2]++;				\
		if (debug_hook > 2)					\
		{							\
		    printf("Command error detected by controller\n\t");	\
		    print_reg(CFSA_ext);				\
		    printf("\tF = %d\n", CFSA_f);			\
		}							\
	    }								\
	    if (ht2992_rlst_derr)					\
	    {								\
		debug_dat.lam_count[3]++;				\
		if (debug_hook > 2)					\
		    printf("Delayed error\n");				\
	    }								\
	    if (ht2992_rlst_adnr)					\
	    {								\
		debug_dat.lam_count[4]++;				\
		if (debug_hook > 2)					\
		{							\
		    printf("Crate Address Not Recognized\n\t");		\
		    print_reg(CFSA_ext);				\
		}							\
	    }								\
	    if (ht2992_rlst_lpe)					\
	    {								\
		debug_dat.lam_count[5]++;				\
		if (debug_hook > 2)					\
		    printf("Longitudinal Parity Error\n");		\
	    }								\
	    if (ht2992_rlst_tmo)					\
	    {								\
		debug_dat.lam_count[6]++;				\
		if (debug_hook > 2)					\
		{							\
		    printf("Timeout on Serial Highway\n\t");		\
		    print_reg(CFSA_ext);				\
		    printf("\tF = %d\n", CFSA_f);			\
		}							\
	    }								\
	    if (ht2992_rlst_tpe)					\
	    {								\
		debug_dat.lam_count[7]++;				\
		if (debug_hook > 2)					\
		    printf("Transverse Parity Error\n");		\
	    }								\
	    if (ht2992_rlst_serr)					\
	    {								\
		debug_dat.lam_count[11]++;				\
		if (debug_hook > 2)					\
		    printf("Lost synchronization error\n");		\
	    }								\
	}								\
	CFSA_try++;							\
							    		\
	/* Clear error register on SHLD				*/	\
	ht2992_wcll = 0x0;						\
							    		\
	/* If this is a read, issue re-read to controller	*/	\
	if ((CFSA_f & 0x0018) == 0x0000)				\
	{								\
									\
	    CFSA_read_try++;						\
									\
	    /* Setup crate address register			*/	\
	    ht2992_wtca = *((short *)(&CFSA_ext));			\
									\
	    CFSA_temp = 0;						\
	    /* pack n and a into lower 16 bits of reg	 	*/	\
	    CFSA_temp |= ((30 << naf_n_shft) & naf_n_mask);		\
	    CFSA_temp |= ((1 << naf_a_shft) & naf_a_mask);		\
									\
	    /* Write NAF (commit for reread)			*/	\
	    ht2992_wnaf = CFSA_temp;					\
	}								\
	else	/* This is a write or a command			*/	\
	{								\
	    /* Setup NAF (commit for CMD)			*/	\
	    ht2992_wnaf = (*((short *)(&CFSA_ext) + 1) | CFSA_f);	\
									\
	    /* If this is a write (xxx10xxx pattern)		*/	\
	    if ((CFSA_f & 0x0018) == 0x0010)				\
	    {								\
		ht2992_wtdt = *(short *)(CFSA_dat);			\
		ht2992_wtdb = *((short *)(CFSA_dat) + 1);		\
		CFSA_write_try++;					\
	    }								\
	    else							\
	    {								\
		CFSA_cmd_try++;						\
	    }								\
	}								\
									\
	/* Spin on status register until reply or timeout	*/	\
	SPINWAIT							\
									\
    }									\
									\
    if (debug_hook)							\
    {									\
	if ((CFSA_f & 0x0018) == 0x0000)	/* Read		*/	\
	    debug_dat.read_error[CFSA_read_try]++;			\
	else if ((CFSA_f & 0x0018) == 0x0010)	/* Write	*/	\
	    debug_dat.write_error[CFSA_write_try]++;			\
	else					/* Command	*/	\
	    debug_dat.cmd_error[CFSA_cmd_try]++;			\
									\
	debug_dat.total++;						\
	debug_dat.total_err += CFSA_try;				\
    }									\
									\
									\
    /* Pass out Q result of transaction				*/	\
    if (ht2992_rcsr & 0x0400)						\
    {									\
	*CFSA_q = 0;							\
    }									\
    else								\
    {									\
	*CFSA_q = ht2992_rcsr_qrsp;					\
									\
	/* If this is a read, unload data			*/	\
	if (CFSA_f < 8)							\
	{								\
	    CFSA_temp2 = ht2992_rddt;/* Top Byte		*/	\
	    CFSA_temp1 = ht2992_rddb;/* Lower 16 bits		*/	\
									\
	    /* Pack 24 bit result into 32 bit buffer		*/	\
	    *(short *)(CFSA_dat) = CFSA_temp2 & 0xff;			\
	    *(((short *)(CFSA_dat)) + 1) = CFSA_temp1;			\
	}								\
    }									\
									\
}


#define CFSA_post(CFSA_camac_lock)					\
{									\
    /* Release lock on interrupts				*/	\
    intUnlock(CFSA_camac_lock);						\
}


#define		CFSA(CFSA1_f, CFSA1_ext, CFSA1_dat, CFSA1_q)		\
{									\
	int	CFSA1_camac_lock;					\
									\
	CFSA_pre(CFSA1_f, CFSA1_camac_lock) ;				\
	CFSA_main(CFSA1_f, CFSA1_ext, CFSA1_dat, CFSA1_q) ;		\
	CFSA_post(CFSA1_camac_lock);					\
									\
}


#define		CFUBC_pre(CFUBC_f, CFUBC_camac_lock)			\
{									\
									\
	if ((CFUBC_f < 0) || (CFUBC_f >= 32))				\
	{								\
		return;							\
	}								\
	else								\
	{								\
		/* Lock out interrupts in critical section 	*/	\
		CFUBC_camac_lock = intLock();				\
	}								\
									\
}									\


#define		CFUBC_main(CFUBC_f, CFUBC_ext, CFUBC_intc, CFUBC_cb)	\
{									\
    unsigned short	CFUBC_temp, CFUBC_temp1, CFUBC_temp2;		\
    int	CFUBC_try;							\
    int CFUBC_count;							\
    int	MAX_TRIES	= 3;						\
									\
    /* Setup crate address register				*/	\
    ht2992_wtca = *((short *)(&CFUBC_ext));				\
							    		\
    /* If this is a write (xxx10xxx pattern)			*/	\
    if ((CFUBC_f & 0x0018) == 0x0010)					\
    {									\
	/* Setup NAF						*/	\
	ht2992_wnaf = (*((short *)(&CFUBC_ext) + 1) | CFUBC_f);		\
									\
	for (CFUBC_count = 0; CFUBC_count < CFUBC_cb[0]; CFUBC_count++)	\
	{								\
	    ht2992_wtdt = *(short *)(CFUBC_intc);			\
	    ht2992_wtdb = *((short *)(CFUBC_intc) + 1);			\
									\
	    /* Spin on status until reply or timeout received	*/	\
	    SPINWAIT							\
									\
	    if (!ht2992_rcsr_qrsp)					\
	    	break;							\
									\
	    CFUBC_cb[1]++;						\
	    CFUBC_intc++;						\
	}								\
									\
    }									\
    else if (CFUBC_f < 8)	/* Must be a read!		*/	\
    {									\
	/* Setup NAF (commit for READ)				*/	\
	ht2992_wnaf = (*((short *)(&CFUBC_ext) + 1) |	 		\
		       CFUBC_f | ht2992_rpt_flag);			\
									\
	for (CFUBC_count = 0; CFUBC_count < CFUBC_cb[0]; CFUBC_count++)	\
	{								\
	    /* Spin on status until reply or timeout received	*/	\
	    SPINWAIT							\
									\
	    if (!ht2992_rcsr_qrsp)					\
	    	break;							\
									\
	    CFUBC_temp2 = ht2992_rddt;/* Top Byte		*/	\
	    CFUBC_temp1 = ht2992_rddb;/* Lower 16 bits		*/	\
									\
	    /* Unload 24 bit result into 32 bit buffer		*/	\
	    *(short *)(CFUBC_intc) = CFUBC_temp2 & 0xff;		\
	    *(((short *)(CFUBC_intc)) + 1) = CFUBC_temp1;		\
									\
	    CFUBC_cb[1]++;						\
	    CFUBC_intc++;						\
									\
	}								\
									\
    }									\
    else	/* Must be a command				*/	\
    {									\
	for (CFUBC_count = 0; CFUBC_count < CFUBC_cb[0]; CFUBC_count++)	\
	{								\
	    /* Setup NAF (commit for CMD)			*/	\
	    ht2992_wnaf = (*((short *)(&CFUBC_ext) + 1) | CFUBC_f );	\
									\
	    /* Spin on status until reply or timeout received	*/	\
	    SPINWAIT							\
									\
	    if (!ht2992_rcsr_qrsp)					\
	    	break;							\
									\
	    CFUBC_cb[1]++;						\
	}								\
    }									\
							    		\
}

#define CFUBC_post(CFUBC_camac_lock)					\
{									\
    /* Release lock on interrupts				*/	\
    intUnlock(CFUBC_camac_lock);					\
}


#define		CFUBC(CFUBC1_f, CFUBC1_ext, CFUBC1_dat, CFUBC1_cb)	\
{									\
	int	CFUBC_camac_lock;					\
									\
	CFUBC_pre(CFUBC1_f, CFUBC_camac_lock) ;				\
	CFUBC_main(CFUBC1_f, CFUBC1_ext, CFUBC1_dat, CFUBC1_cb) ;	\
	CFUBC_post(CFUBC_camac_lock);					\
									\
}


#define		CSSA_pre(CSSA_f, CSSA_camac_lock)			\
{									\
									\
	if ((CSSA_f < 0) || (CSSA_f >= 32))				\
	{								\
		return;							\
	}								\
	else								\
	{								\
		/* Lock out interrupts in critical section 	*/	\
		CSSA_camac_lock = intLock();				\
	}								\
									\
}									\


#define		CSSA_main(CSSA_f, CSSA_ext, CSSA_dat, CSSA_q)		\
{									\
    unsigned short	CSSA_temp, CSSA_temp1, CSSA_temp2;		\
    int	CSSA_try, CSSA_read_try, CSSA_write_try, CSSA_cmd_try;		\
    int	MAX_TRIES	= 3;						\
    int	b_temp, c_temp, n_temp, a_temp;					\
									\
    /* Setup crate address register				*/	\
    ht2992_wtca = *((short *)(&CSSA_ext));				\
							    		\
    /* Setup NAF (commit for RD/CMD)				*/	\
    ht2992_wnaf = (*((short *)(&CSSA_ext) + 1) | CSSA_f);		\
							    		\
    /* If this is a write (xxx10xxx pattern)			*/	\
    if ((CSSA_f & 0x0018) == 0x0010)					\
    {									\
	ht2992_wtdb = *CSSA_dat;					\
    }									\
							    		\
    /* Spin on status bit until reply or timeout received	*/	\
    SPINWAIT								\
									\
									\
    /* If timeout or other error set, clear error flag		*/	\
    /*     and do "re-read" of L2 controller			*/	\
    CSSA_try = CSSA_read_try = CSSA_write_try = CSSA_cmd_try = 0;	\
    while ((ht2992_rcsr & 0x0400) && (CSSA_try < MAX_TRIES))		\
    {									\
	if (debug_hook)							\
	{								\
	    if (ht2992_rlst_sq == 1)					\
	    {								\
		debug_dat.lam_count[0]++;				\
		if (debug_hook > 1)					\
		{							\
		    printf("No Q returned from last CSSA\n");		\
		    print_reg(CSSA_ext);				\
		}							\
	    }								\
	    if (ht2992_rlst_sx)						\
	    {								\
		debug_dat.lam_count[1]++;				\
		if (debug_hook > 1)					\
		{							\
		    printf("No X Returned from last CSSA\n");		\
		    print_reg(CSSA_ext);				\
		}							\
	    }								\
	    if (ht2992_rlst_err)					\
	    {								\
		debug_dat.lam_count[2]++;				\
		if (debug_hook > 2)					\
		{							\
		    printf("Command error detected by controller\n\t");	\
		    print_reg(CSSA_ext);				\
		    printf("\tF = %d\n", CSSA_f);			\
		}							\
	    }								\
	    if (ht2992_rlst_derr)					\
	    {								\
		debug_dat.lam_count[3]++;				\
		if (debug_hook > 2)					\
		    printf("Delayed error\n");				\
	    }								\
	    if (ht2992_rlst_adnr)					\
	    {								\
		debug_dat.lam_count[4]++;				\
		if (debug_hook > 2)					\
		{							\
		    printf("Crate Address Not Recognized\n\t");		\
		    print_reg(CSSA_ext);				\
		}							\
	    }								\
	    if (ht2992_rlst_lpe)					\
	    {								\
		debug_dat.lam_count[5]++;				\
		if (debug_hook > 2)					\
		    printf("Longitudinal Parity Error\n");		\
	    }								\
	    if (ht2992_rlst_tmo)					\
	    {								\
		debug_dat.lam_count[6]++;				\
		if (debug_hook > 2)					\
		{							\
		    printf("Timeout on Serial Highway\n\t");		\
		    print_reg(CSSA_ext);				\
		    printf("\tF = %d\n", CSSA_f);			\
		}							\
	    }								\
	    if (ht2992_rlst_tpe)					\
	    {								\
		debug_dat.lam_count[7]++;				\
		if (debug_hook > 2)					\
		    printf("Transverse Parity Error\n");		\
	    }								\
	    if (ht2992_rlst_serr)					\
	    {								\
		debug_dat.lam_count[11]++;				\
		if (debug_hook > 2)					\
		    printf("Lost synchronization error\n");		\
	    }								\
	}								\
	CSSA_try++;							\
							    		\
	/* Clear error register on SHLD				*/	\
	ht2992_wcll = 0x0;						\
							    		\
	/* If this is a read, issue re-read to controller	*/	\
	if ((CSSA_f & 0x0018) == 0x0000)				\
	{								\
	    CSSA_read_try++;						\
									\
	    /* Setup crate address register			*/	\
	    ht2992_wtca = *((short *)(&CSSA_ext));			\
									\
	    CSSA_temp = 0;						\
	    /* pack n and a into lower 16 bits of reg	 	*/	\
	    CSSA_temp |= ((30 << naf_n_shft) & naf_n_mask);		\
	    CSSA_temp |= ((1 << naf_a_shft) & naf_a_mask);		\
									\
	    /* Write NAF (commit for reread)			*/	\
	    ht2992_wnaf = CSSA_temp;					\
	}								\
	else	/* This is a write or a command			*/	\
	{								\
	    /* Setup NAF (commit for CMD)			*/	\
	    ht2992_wnaf = (*((short *)(&CSSA_ext) + 1) | CSSA_f);	\
									\
	    /* If this is a write (xxx10xxx pattern)		*/	\
	    if ((CSSA_f & 0x0018) == 0x0010)				\
	    {								\
		ht2992_wtdb = *CSSA_dat;				\
		CSSA_write_try++;					\
	    }								\
	    else							\
	    {								\
		CSSA_cmd_try++;						\
	    }								\
	}								\
									\
	/* Spin on status register until reply or timeout	*/	\
	SPINWAIT							\
									\
    }									\
									\
    if (debug_hook)							\
    {									\
	if ((CSSA_f & 0x0018) == 0x0000)	/* Read		*/	\
	    debug_dat.read_error[CSSA_read_try]++;			\
	else if ((CSSA_f & 0x0018) == 0x0010)	/* Write	*/	\
	    debug_dat.write_error[CSSA_write_try]++;			\
	else					/* Command	*/	\
	    debug_dat.cmd_error[CSSA_cmd_try]++;			\
									\
	debug_dat.total++;						\
	debug_dat.total_err += CSSA_try;				\
    }									\
									\
									\
    /* Pass out Q result of transaction				*/	\
    if (ht2992_rcsr & 0x0400)						\
    {									\
	*CSSA_q = 0;							\
    }									\
    else								\
    {									\
	*CSSA_q = ht2992_rcsr_qrsp;					\
									\
	/* If this is a read, unload data			*/	\
	if (CSSA_f < 8)							\
	{								\
	    *CSSA_dat = ht2992_rddb;/* Lower 16 bits		*/	\
	}								\
    }									\
									\
}


#define CSSA_post(CSSA_camac_lock)					\
{									\
    /* Release lock on interrupts				*/	\
    intUnlock(CSSA_camac_lock);						\
}


#define		CSSA(CSSA1_f, CSSA1_ext, CSSA1_dat, CSSA1_q)		\
{									\
	int	CSSA1_camac_lock;					\
									\
	CSSA_pre(CSSA1_f, CSSA1_camac_lock) ;				\
	CSSA_main(CSSA1_f, CSSA1_ext, CSSA1_dat, CSSA1_q) ;		\
	CSSA_post(CSSA1_camac_lock);					\
									\
}
#define		CFUBR_pre(CFUBR_f, CFUBR_camac_lock)			\
{									\
									\
	if ((CFUBR_f < 0) || (CFUBR_f >= 32))				\
	{								\
		return;							\
	}								\
	else								\
	{								\
		/* Lock out interrupts in critical section 	*/	\
		CFUBR_camac_lock = intLock();				\
	}								\
									\
}									\


#define		CFUBR_main(CFUBR_f, CFUBR_ext, CFUBR_intc, CFUBR_cb)	\
{									\
    unsigned short	CFUBR_temp, CFUBR_temp1, CFUBR_temp2;		\
    int	CFUBR_try;							\
    int CFUBR_count;							\
    int	MAX_TRIES	= 3;						\
									\
    /* Setup crate address register				*/	\
    ht2992_wtca = *((short *)(&CFUBR_ext));				\
							    		\
    /* If this is a write (xxx10xxx pattern)			*/	\
    if ((CFUBR_f & 0x0018) == 0x0010)					\
    {									\
	/* Setup NAF						*/	\
	ht2992_wnaf = (*((short *)(&CFUBR_ext) + 1) | CFUBR_f);		\
									\
	while (CFUBR_cb[1] < CFUBR_cb[0])				\
	{								\
	    for (CFUBR_try = 0; CFUBR_try < MAX_TRIES; CFUBR_try++)	\
	    {								\
		ht2992_wtdt = *(short *)(CFUBR_intc);			\
		ht2992_wtdb = *((short *)(CFUBR_intc) + 1);		\
									\
		/* Spin on status until reply or timeout received*/	\
		SPINWAIT						\
									\
		if (ht2992_rcsr_qrsp)	/* if Q response rcvd	*/	\
		{							\
		    CFUBR_cb[1]++;	/* Increment count	*/	\
		    break;		/* Go to next op.	*/	\
		}							\
	    }								\
	}								\
									\
    }									\
    else if (CFUBR_f < 8)	/* Must be a read!		*/	\
    {									\
	while (CFUBR_cb[1] < CFUBR_cb[0])				\
	{								\
	    for (CFUBR_try = 0; CFUBR_try < MAX_TRIES; CFUBR_try++)	\
	    {								\
		/* Setup NAF (commit for READ)			*/	\
		ht2992_wnaf = (*((short *)(&CFUBR_ext) + 1) | 		\
			       CFUBR_f ); 				\
									\
		/* Spin on status until reply or timeout received*/	\
		SPINWAIT						\
									\
		if (ht2992_rcsr_qrsp)	/* if Q response rcvd	*/	\
		{							\
		    CFUBR_temp2 = ht2992_rddt;/* Top Byte	*/	\
		    CFUBR_temp1 = ht2992_rddb;/* Lower 16 bits	*/	\
									\
		    /* Unload 24 bit result into 32 bit buffer	*/	\
		    *(short *)(CFUBR_intc) = CFUBR_temp2 & 0xff;	\
		    *(((short *)(CFUBR_intc)) + 1) = CFUBR_temp1;	\
									\
		    CFUBR_cb[1]++;	/* Increment count	*/	\
		    CFUBR_intc++;	/* Increment data pointer */	\
									\
		    break;		/* Go to next op.	*/	\
		}							\
	    }								\
	}								\
    }									\
    else	/* Must be a command				*/	\
    {									\
	while (CFUBR_cb[1] < CFUBR_cb[0])				\
	{								\
	    for (CFUBR_try = 0; CFUBR_try < MAX_TRIES; CFUBR_try++)	\
	    {								\
		/* Setup NAF (commit for CMD)			*/	\
		ht2992_wnaf = (*((short *)(&CFUBR_ext) + 1) | CFUBR_f );\
									\
		/* Spin on status until reply or timeout rcvd	*/	\
		SPINWAIT						\
									\
		if (ht2992_rcsr_qrsp)					\
		{							\
		    cb[1]++;						\
		    break;						\
		}							\
	    }								\
	}								\
    }									\
							    		\
}

#define CFUBR_post(CFUBR_camac_lock)					\
{									\
    /* Release lock on interrupts				*/	\
    intUnlock(CFUBR_camac_lock);					\
}


#define		CFUBR(CFUBR1_f, CFUBR1_ext, CFUBR1_dat, CFUBR1_cb)	\
{									\
	int	CFUBR_camac_lock;					\
									\
	CFUBR_pre(CFUBR1_f, CFUBR_camac_lock) ;				\
	CFUBR_main(CFUBR1_f, CFUBR1_ext, CFUBR1_dat, CFUBR1_cb) ;	\
	CFUBR_post(CFUBR_camac_lock);					\
									\
}

#define		CSUBR_pre(CSUBR_f, CSUBR_camac_lock)			\
{									\
									\
	if ((CSUBR_f < 0) || (CSUBR_f >= 32))				\
	{								\
		return;							\
	}								\
	else								\
	{								\
		/* Lock out interrupts in critical section 	*/	\
		CSUBR_camac_lock = intLock();				\
	}								\
									\
}									\


#define		CSUBR_main(CSUBR_f, CSUBR_ext, CSUBR_intc, CSUBR_cb)	\
{									\
    unsigned short	CSUBR_temp, CSUBR_temp1, CSUBR_temp2;		\
    int	CSUBR_try;							\
    int CSUBR_count;							\
    int	MAX_TRIES	= 3;						\
									\
    /* Setup crate address register				*/	\
    ht2992_wtca = *((short *)(&CSUBR_ext));				\
							    		\
    /* If this is a write (xxx10xxx pattern)			*/	\
    if ((CSUBR_f & 0x0018) == 0x0010)					\
    {									\
	/* Setup NAF						*/	\
	ht2992_wnaf = (*((short *)(&CSUBR_ext) + 1) | CSUBR_f);		\
									\
	while (CSUBR_cb[1] < CSUBR_cb[0])				\
	{								\
	    for (CSUBR_try = 0; CSUBR_try < MAX_TRIES; CSUBR_try++)	\
	    {								\
		ht2992_wtdb = *CSUBR_intc;				\
									\
		/* Spin on status until reply or timeout received*/	\
		SPINWAIT						\
									\
		if (ht2992_rcsr_qrsp)	/* if Q response rcvd	*/	\
		{							\
		    CSUBR_cb[1]++;	/* Increment count	*/	\
		    break;		/* Go to next op.	*/	\
		}							\
	    }								\
	}								\
									\
    }									\
    else if (CSUBR_f < 8)	/* Must be a read!		*/	\
    {									\
	while (CSUBR_cb[1] < CSUBR_cb[0])				\
	{								\
	    for (CSUBR_try = 0; CSUBR_try < MAX_TRIES; CSUBR_try++)	\
	    {								\
		/* Setup NAF (commit for READ)			*/	\
		ht2992_wnaf = (*((short *)(&CSUBR_ext) + 1) | 		\
			       CSUBR_f ); 				\
									\
		/* Spin on status until reply or timeout received*/	\
		SPINWAIT						\
									\
		if (ht2992_rcsr_qrsp)	/* if Q response rcvd	*/	\
		{							\
									\
		    /* Unload 16 bit result into 16 bit buffer	*/	\
		    *CSUBR_intc = ht2992_rddb;/* Lower 16 bits	*/	\
									\
		    CSUBR_cb[1]++;	/* Increment count	*/	\
		    CSUBR_intc++;	/* Increment data pointer */	\
									\
		    break;		/* Go to next op.	*/	\
		}							\
	    }								\
	}								\
    }									\
    else	/* Must be a command				*/	\
    {									\
	while (CSUBR_cb[1] < CSUBR_cb[0])				\
	{								\
	    for (CSUBR_try = 0; CSUBR_try < MAX_TRIES; CSUBR_try++)	\
	    {								\
		/* Setup NAF (commit for CMD)			*/	\
		ht2992_wnaf = (*((short *)(&CSUBR_ext) + 1) | CSUBR_f );\
									\
		/* Spin on status until reply or timeout rcvd	*/	\
		SPINWAIT						\
									\
		if (ht2992_rcsr_qrsp)					\
		{							\
		    CSUBR_cb[1]++;					\
		    break;						\
		}							\
	    }								\
	}								\
    }									\
							    		\
}

#define CSUBR_post(CSUBR_camac_lock)					\
{									\
    /* Release lock on interrupts				*/	\
    intUnlock(CSUBR_camac_lock);					\
}


#define		CSUBR(CSUBR1_f, CSUBR1_ext, CSUBR1_dat, CSUBR1_cb)	\
{									\
	int	CSUBR_camac_lock;					\
									\
	CSUBR_pre(CSUBR1_f, CSUBR_camac_lock) ;				\
	CSUBR_main(CSUBR1_f, CSUBR1_ext, CSUBR1_dat, CSUBR1_cb) ;	\
	CSUBR_post(CSUBR_camac_lock);					\
									\
}

#define		CSUBC_pre(CSUBC_f, CSUBC_camac_lock)			\
{									\
									\
	if ((CSUBC_f < 0) || (CSUBC_f >= 32))				\
	{								\
		return;							\
	}								\
	else								\
	{								\
		/* Lock out interrupts in critical section 	*/	\
		CSUBC_camac_lock = intLock();				\
	}								\
									\
}									\


#define		CSUBC_main(CSUBC_f, CSUBC_ext, CSUBC_intc, CSUBC_cb)	\
{									\
    unsigned short	CSUBC_temp, CSUBC_temp1, CSUBC_temp2;		\
    int	CSUBC_try;							\
    int CSUBC_count;							\
    int	MAX_TRIES	= 3;						\
									\
    /* Setup crate address register				*/	\
    ht2992_wtca = *((short *)(&CSUBC_ext));				\
							    		\
    /* If this is a write (xxx10xxx pattern)			*/	\
    if ((CSUBC_f & 0x0018) == 0x0010)					\
    {									\
	/* Setup NAF						*/	\
	ht2992_wnaf = (*((short *)(&CSUBC_ext) + 1) | CSUBC_f);		\
									\
	for (CSUBC_count = 0; CSUBC_count < CSUBC_cb[0]; CSUBC_count++)	\
	{								\
	    ht2992_wtdb = *CSUBC_intc;					\
									\
	    /* Spin on status until reply or timeout received	*/	\
	    SPINWAIT							\
									\
	    if (!ht2992_rcsr_qrsp)					\
	    	break;							\
									\
	    CSUBC_cb[1]++;						\
	    CSUBC_intc++;						\
	}								\
									\
    }									\
    else if (CSUBC_f < 8)	/* Must be a read!		*/	\
    {									\
	/* Setup NAF (commit for READ)				*/	\
	ht2992_wnaf = (*((short *)(&CSUBC_ext) + 1) |	 		\
		       CSUBC_f | 					\
		       ht2992_rpt_flag);				\
									\
	for (CSUBC_count = 0; CSUBC_count < CSUBC_cb[0]; CSUBC_count++)	\
	{								\
	    /* Spin on status until reply or timeout received	*/	\
	    SPINWAIT							\
									\
	    if (!ht2992_rcsr_qrsp)					\
	    	break;							\
									\
	    *CSUBC_intc = ht2992_rddb;/* Lower 16 bits		*/	\
									\
	    CSUBC_cb[1]++;						\
	    CSUBC_intc++;						\
									\
	}								\
									\
    }									\
    else	/* Must be a command				*/	\
    {									\
	for (CSUBC_count = 0; CSUBC_count < CSUBC_cb[0]; CSUBC_count++)	\
	{								\
	    /* Setup NAF (commit for CMD)			*/	\
	    ht2992_wnaf = (*((short *)(&CSUBC_ext) + 1) | CSUBC_f );	\
									\
	    /* Spin on status until reply or timeout received	*/	\
	    SPINWAIT							\
									\
	    if (!ht2992_rcsr_qrsp)					\
	    	break;							\
									\
	    CSUBC_cb[1]++;						\
	}								\
    }									\
							    		\
}

#define CSUBC_post(CSUBC_camac_lock)					\
{									\
    /* Release lock on interrupts				*/	\
    intUnlock(CSUBC_camac_lock);					\
}


#define		CSUBC(CSUBC1_f, CSUBC1_ext, CSUBC1_dat, CSUBC1_cb)	\
{									\
	int	CSUBC_camac_lock;					\
									\
	CSUBC_pre(CSUBC1_f, CSUBC_camac_lock) ;				\
	CSUBC_main(CSUBC1_f, CSUBC1_ext, CSUBC1_dat, CSUBC1_cb) ;	\
	CSUBC_post(CSUBC_camac_lock);					\
									\
}

